package io.tswf.minecraft.blockventure.core.transmutation.transmute

import groovy.transform.CompileStatic
import groovy.util.logging.Slf4j
import io.tswf.minecraft.blockventure.core.exception.ServerWorldCustomStoragesSaveException
import io.tswf.minecraft.blockventure.core.transmutation.expansion.IServerWorldCustomDataStorageAccess
import io.tswf.minecraft.blockventure.core.world.chunk.persist.ChunkPersistService
import io.tswf.transmutations.common.annotation.ClassReference
import io.tswf.transmutations.common.annotation.Constant
import io.tswf.transmutations.common.annotation.MethodDescription
import io.tswf.transmutations.transmute.annotation.AtHead
import io.tswf.transmutations.transmute.annotation.ReturningConstantIfTrue
import io.tswf.transmutations.transmute.annotation.Self
import io.tswf.transmutations.transmute.annotation.Transmute
import io.tswf.transmutations.transmute.meta.MissingTargetStrategy
import net.minecraft.server.world.ServerWorld
import net.minecraft.util.ProgressListener
import net.minecraft.util.math.ChunkPos

@Slf4j
@CompileStatic
class TransmutationServerWorld {

    static boolean logStoragesSaving = System.findGlobalProperty("blockventure.world.custom-storage.saving.logging.enabled")
        .map { it.toBoolean() }
        .orElse(true)

    @Transmute(
            targetClass = @ClassReference(ServerWorld),
            targetMethod = @MethodDescription(
                    name = "save",
                    parameterTypeClasses = [ ProgressListener, ClassReference.BOOLEAN, ClassReference.BOOLEAN ],
                    returnTypeClass = ClassReference.VOID
            ),
            onMissing = MissingTargetStrategy.THROW
    )
    @AtHead
    static void handleServerWorldSave(ProgressListener progressListener, boolean flush, boolean savingDisabled, @Self ServerWorld world) {
        try {
            def storageAccess = (IServerWorldCustomDataStorageAccess) world
            def storages = storageAccess.blockventureCustomStorages

            for (def storage : storages.values()) {
                if (logStoragesSaving) {
                    log.info("Saving blockventure custom storage named '${storage.name}' [${world.dimensionKey.value}]...")
                }

                storage.commit()

                if (logStoragesSaving) {
                    log.info("Blockventure custom storage named '${storage.name}' [${world.dimensionKey.value}] saved!")
                }
            }
        } catch (Throwable e) {
            if (e instanceof ServerWorldCustomStoragesSaveException.Bypass) {
                throw e
            }

            throw new ServerWorldCustomStoragesSaveException(world, e)
        }
    }

    @Transmute(
            targetClass = @ClassReference(ServerWorld),
            targetMethod = @MethodDescription(
                    name = "shouldTickBlocksInChunk",
                    parameterTypeClasses = [ ClassReference.LONG ],
                    returnTypeClass = ClassReference.BOOLEAN
            ),
            onMissing = MissingTargetStrategy.THROW
    )
    @AtHead
    @ReturningConstantIfTrue(@Constant(booleanValue = true))
    static boolean shouldForceTickBlockChunk(long chunkPos, @Self ServerWorld serverWorld) {
        return ChunkPersistService.isChunkPersisted(serverWorld, chunkPos)
    }
}
